/**
* avl.h -- public types and external declarations for avl trees
*
* Created 03/01/89 by Brad Appleton
*
* Fri Jul 14 13:54:12 1989, Rev 1.0, brad(0165)
*
* Hacked for PSPP by BLP 7/21/95, 8/12/95.
*
**/

#if !avl_h
#define avl_h 1

#include "arena.h"

/* Definition of traversal type. */
typedef enum
  {
    PREORDER,
    INORDER,
    POSTORDER,
  }
VISIT;

/* Definition of sibling order type. */
typedef enum
  {
    LEFT_TO_RIGHT,
    RIGHT_TO_LEFT,
  }
SIBLING_ORDER;

/* Definition of node type. */
typedef enum
  {
    IS_TREE,			/* Both subtrees are non-empty. */
    IS_LBRANCH,			/* Left subtree is non-empty; right is empty. */
    IS_RBRANCH,			/* Right subtree is non-empty; left is empty. */
    IS_LEAF,			/* Both subtrees are empty. */
    IS_NULL,			/* Tree is empty. */
  }
NODE;

/* Structure for a node in an AVL tree. */
typedef struct avl_node
  {
    void *data;			/* Pointer to data. */
    short bal;			/* Balance factor. */
    struct avl_node *subtree[2];	/* LEFT and RIGHT subtrees. */
  }
avl_node;

/* Used for traversing an AVL tree. */
typedef struct
  {
    int nstack;			/* Top of stack. */
    avl_node *p;		/* Used for traversal. */
    avl_node *stack[48];	/* Descended trees (up to 2**32 nodes!). */
  }
avl_traverser;

/* Function types. */
typedef int (*avl_comparison_func) (void *a, void *b, void *param);
typedef void (*avl_free_func) (void *data, void *param);
typedef void (*avl_walk_func) (void *data, void *param, int level);

/* Structure which holds information about an AVL tree. */
typedef struct
  {
    arena **owner;		/* Arena to store nodes. */
    avl_node *root;		/* Root node of the tree. */
    avl_comparison_func compare;	/* Used to compare keys. */
    int count;			/* Number of nodes in the tree. */
    void *param;		/* Arbitary user data. */
  }
avl_tree;

/* General functions. */
avl_tree *avl_create (arena **, avl_comparison_func, void *param);
void avl_destroy (avl_tree *, avl_free_func);
void avl_free (avl_tree *);
int avl_count (avl_tree *);
avl_tree *avl_sort (avl_tree *, arena **, avl_comparison_func, void *param);
avl_tree *avl_shallow_copy (avl_tree *);

/* Walk the tree. */
void avl_walk_preorder (avl_tree *, avl_walk_func, void *param);
void avl_walk_inorder (avl_tree *, avl_walk_func, void *param);
void *avl_traverse (avl_tree *, avl_traverser **);

/* Search for a given item. */
void *avl_insert (avl_tree *, void *);
void *avl_replace (avl_tree *, void *);
void *avl_delete (avl_tree *, void *);
void *avl_find (avl_tree *, void *);

/* Search for the minimal item. */
void *avl_delete_min (avl_tree *);
void *avl_find_min (avl_tree *);

/* Search for the maximal item. */
void *avl_delete_max (avl_tree *);
void *avl_find_max (avl_tree *);

/* Easy assertions on insertion & deletion. */
#if GLOBAL_DEBUGGING
#define force_avl_insert(A, B)			\
	do					\
	  {					\
	    if (avl_insert (A, B))		\
	      assert(0);			\
	  }					\
	while (0)
void *force_avl_delete (avl_tree *, void *);
#else
#define force_avl_insert(A, B) 			\
	avl_insert (A, B)
#define force_avl_delete(A, B) 			\
	avl_delete (A, B)
#endif

#endif /* avl_h */
